@@ -63,6 +63,7 @@ class SASconfigHTTP:
       self.authkey   = cfg.get('authkey', '')
       self._prompt   = session._sb.sascfg._prompt
       self.lrecl     = cfg.get('lrecl', None)
+      self.inactive  = cfg.get('inactive', 120)
 
       try:
          self.outopts = getattr(SAScfg, "SAS_output_options")
@@ -188,6 +189,13 @@ class SASconfigHTTP:
       if not self.lrecl:
          self.lrecl = 1048576
 
+      inito = kwargs.get('inactive', None)
+      if inito is not None:
+         if lock and self.inactive:
+            print("Parameter 'inactive' passed to SAS_session was ignored due to configuration restriction.")
+         else:
+            self.inactive = inito
+
       inak = kwargs.get('authkey', '')
       if len(inak) > 0:
          if lock and len(self.authkey):
@@ -263,9 +271,11 @@ class SASconfigHTTP:
          code_pw = ''
          if len(user) == 0:
             msg  = "To connect to Viya you need either an authcode or a userid/pw. Neither were provided.\n"
-            msg += "Please enter which one you want to enter next. Type one of these now: [authcode | userid]: " 
+            msg += "Please enter which one you want to enter next. Type one of these now: [default=authcode | userid]: " 
             while code_pw.lower() not in ['userid','authcode']:
                code_pw = self._prompt(msg)
+               if code_pw == '':
+                  code_pw = 'authcode'
                if code_pw is None:
                   self._token = None
                   raise RuntimeError("Neither authcode nor userid provided.") 
@@ -288,12 +298,12 @@ class SASconfigHTTP:
                if user is None:
                   self._token = None
                   raise RuntimeError("No userid provided.") 
-        
+      
             while len(pw) == 0:
                pw = self._prompt("Please enter password: ", pw = True)
                if pw is None:
                   self._token = None
-                  raise RuntimeError("No password provided.") 
+                  raise RuntimeError("No password provided.")
 
       if self.ssl:
          if self.verify:
@@ -539,8 +549,10 @@ class SASsessionHTTP():
          "POST uri not found in context info. You may not have permission to use this context.\n{}".format(self.sascfg.ctx))
 
       conn = self.sascfg.HTTPConn; conn.connect()
-      d1 = '{"name":"'+self.sascfg.ctxname+'", "description":"saspy session", "version":1, "environment":{"options":'+options+'}}'
-      headers={"Accept":"application/vnd.sas.compute.session+json","Content-Type":"application/vnd.sas.compute.session.request+json","Authorization":"Bearer "+self.sascfg._token}
+      d1  = '{"name":"'+self.sascfg.ctxname+'", "description":"saspy session", "version":1, "environment":{"options":'+options+'}'
+      d1 += ',"attributes": {"sessionInactiveTimeout": '+str(int(float(self.sascfg.inactive)*60))+'}}'
+      headers={"Accept":"application/vnd.sas.compute.session+json","Content-Type":"application/vnd.sas.compute.session.request+json",
+               "Authorization":"Bearer "+self.sascfg._token}
 
       try:
          conn.request('POST', uri, body=d1, headers=headers)
